// General-purpose Photovoltaic Device Model gpvdm.com - a drift diffusion
// base/Shockley-Read-Hall model for 1st, 2nd and 3rd generation solarcells.
// The model can simulate OLEDs, Perovskite cells, and OFETs.
// 
// Copyright (C) 2008-2020 Roderick C. I. MacKenzie
// All rights reserved.
// 
// https://www.gpvdm.com
// r.c.i.mackenzie at googlemail.com
// 
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of the GPVDM nor the
//       names of its contributors may be used to endorse or promote products
//       derived from this software without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL Roderick C. I. MacKenzie BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <errno.h>
#include "sim.h"
#include "sun.h"
#include "lock.h"
#include "dump.h"
#include "complex_solver.h"

static double min_light_error=1e-10;

static double last_Psun=-1000.0;
static double last_Plaser=-1000.0;
static double last_wavelength_laser=-1000.0;

void light_set_dx(struct light *in,double dx)
{
in->dx=dx;
}

void light_free(struct light *in)
{
int i;
free(in->H1d);
free(in->thick);
free(in->x);

for (i=0;i<in->layers;i++)
{
	inter_free(&(in->mat[i]));
	inter_free(&(in->mat_n[i]));
}

for (i=0;i<in->lpoints;i++)
{
	free(in->En[i]);
	free(in->Enz[i]);
	free(in->Ep[i]);
	free(in->Epz[i]);

	free(in->photons_asb[i]);
	free(in->alpha[i]);
	free(in->photons[i]);
	free(in->n[i]);
	free(in->t[i]);
	free(in->r[i]);
	free(in->H[i]);
}

free(in->layer_end);
free(in->G_percent);
free(in->enable_gen);
free(in->t);
free(in->r);
free(in->En);
free(in->Enz);
free(in->Ep);
free(in->Epz);
free(in->photons_asb);
free(in->alpha);
free(in->photons);
free(in->n);
free(in->H);

free(in->sun);
free(in->sun_norm);
free(in->sun_photons);
free(in->sun_E);
free(in->Gn);
free(in->Gp);
free(in->layer);
free(in->Ti);
free(in->Tj);
free(in->Tx);
free(in->Txz);
free(in->b);
free(in->bz);
free(in->l);
}

double light_cal_photon_density(struct light *in)
{
int i;
int ii;
double tot=0.0;
double Enl=0.0;
double Epl=0.0;
double Enc=0.0;
double Epc=0.0;
double H_tot=0.0;

for (ii=0;ii<in->points;ii++)
{
	tot=0.0;
	H_tot=0.0;
	for (i=0;i<in->lpoints;i++)
	{
		in->photons[i][ii]=0.5*epsilon0*cl*in->n[i][ii]*(pow(in->Ep[i][ii]+in->En[i][ii],2.0)+pow(in->Enz[i][ii]+in->Epz[i][ii],2.0));
		in->photons[i][ii]*=(in->l[i]/(hp*cl));
		in->photons_asb[i][ii]=in->photons[i][ii]*in->alpha[i][ii];

		double E=((hp*cl)/in->l[i])/q-in->Eg;

		//getchar();
		if (E>0.0)
		{
		in->H[i][ii]=E*q*in->photons_asb[i][ii];
		}else
		{
		in->H[i][ii]=0.0;
		}

		//printf("%d %d %le %le %le\n",i,ii,E,in->H[i][ii],in->photons_asb[i][ii]);

		tot+=in->photons_asb[i][ii]*in->dl;
		H_tot+=in->H[i][ii]*in->dl;
		//printf("%le %le\n",E,in->l[i]);
		//getchar();

	}

in->Gn[ii]=tot;
in->Gp[ii]=tot;
in->H1d[ii]=H_tot;
}

/*FILE *out;
out=fopen("./light_test.dat","w");
for (i=0;i<in->lpoints;i++)
{
	tot=0.0;
	for (ii=0;ii<in->points;ii++)
	{
		fprintf(out,"%le %le %le\n",in->l[i],in->x[ii],in->alpha[i][ii]*(in->sun_photons[i])*(exp(-in->alpha[i][ii]*(in->x[ii]))+exp(-2*in->alpha[i][ii]*in->x[in->points-1])*exp(in->alpha[i][ii]*in->x[ii])));
	}

in->G[ii]=tot;
}
fclose(out);*/




}

void light_set_sun_power(struct light *in,double power, double laser)
{
int i;

double E=0.0;
double tot0=0.0;

for  (i=0;i<in->lpoints;i++)
{
	in->sun[i]=in->sun_norm[i]*power;

	E=hp*cl/in->l[i];
	in->sun_photons[i]=in->sun[i]/E;
	if (i==in->laser_pos)
	{
	//double E_laser=hp*cl/in->laser_wavelength;

	//printf("laser photons %le\n",in->lasereff*((in->pulseJ/in->pulseL/E_laser)/(in->spotx*in->spoty))/in->dl);
	//exit(0);
	if (in->laser_photon_density==-1.0)
	{
	in->sun_photons[i]+=(laser/100.0)*in->lasereff*((in->pulseJ/in->pulseL/E)/(in->spotx*in->spoty))/in->dl;
	//printf("%le\n",laser*in->lasereff*((in->pulseJ/in->pulseL/E)/(in->spotx*in->spoty))/in->dl);
	//getchar();
	}else
	{
	in->sun_photons[i]+=in->laser_photon_density/in->dl;
	}
	}

	in->sun_E[i]=pow(2.0*(in->sun_photons[i]*E)/(epsilon0*cl*in->n[i][0]),0.5);
	//printf("%le\n",in->sun_E[i]);

}

/*for  (i=0;i<in->lpoints;i++)
{

	tot0+=in->sun[i]*in->dl;

}
printf("%le\n",tot0);
exit(0);*/

}

void solve_light(struct device *cell,struct light *in,double Psun_in,double Plaser_in)
{
int i;
int ii;
//dprintf("Solve light\n");

if ((last_Plaser!=Plaser_in)||(last_Psun!=Psun_in)||(last_wavelength_laser!=in->laser_wavelength))
{
printf("Propagating\n");
	last_Plaser=Plaser_in;
	last_Psun=Psun_in;
	last_wavelength_laser=in->laser_wavelength;
	double Psun=Psun_in*pow(10.0,-in->ND);

	light_set_sun_power(in,Psun,Plaser_in);

	if ((Plaser_in==0)&&(Psun_in==0))
	{
		if (get_dump_status(dump_optics)==TRUE) printf("It's dark I know what the answer is\n");
		for (i=0;i<in->lpoints;i++)
		{
			for (ii=0;ii<in->points;ii++)
			{
				in->Ep[i][ii]=0.0;
				in->En[i][ii]=0.0;
				in->Enz[i][ii]=0.0;
				in->Epz[i][ii]=0.0;
			}
		}
	}else
	{
		for (i=0;i<in->lpoints;i++)
		{
			light_solve_lam_slice(in,i);
			//printf("%le %le\n",Psun,in->sun_E[i]);
		}
	}
	light_cal_photon_density(in);
	double Gn=0.0;
	double Gp=0.0;
	double H=0.0;
	double tot=0.0;

	if (in->align_mesh==FALSE)
	{
		for (i=0;i<cell->ymeshpoints;i++)
		{
			Gn=inter_get_raw(in->x,in->Gn,in->points,in->device_start+cell->ymesh[i])*cell->Dphotoneff;
			Gp=inter_get_raw(in->x,in->Gp,in->points,in->device_start+cell->ymesh[i])*cell->Dphotoneff;
			H=inter_get_raw(in->x,in->H1d,in->points,in->device_start+cell->ymesh[i]);


			cell->Gn[i]=Gn*in->electron_eff;
			cell->Gp[i]=Gp*in->hole_eff;
			cell->Habs[i]=H;

		}
	}else
	{
		for (i=0;i<cell->ymeshpoints;i++)
		{
			cell->Gn[i]=in->Gn[in->device_start_i+i]*cell->Dphotoneff;
			cell->Gp[i]=in->Gp[in->device_start_i+i]*cell->Dphotoneff;//changed to Gp
		}

	}


	int stored=FALSE;
	tot=0.0;
	if (in->gather==TRUE)
	{
		for (i=0;i<cell->ymeshpoints;i++)
		{
			if (cell->imat[i]==0)
			{
				if (fabs(cell->ymesh[i]-cell->mat.l[0].height)<in->dlength)
				{
					tot+=cell->Gn[i];
					printf("yes\n");
				}else
				{
					printf("no\n");
				}
				cell->Gn[i]=0.0;
				cell->Gp[i]=0.0;
				stored=TRUE;
			}else
			{
				if (stored==TRUE)
				{
					if ((i-1)>=0)
					{
						cell->Gp[i-1]=tot;
					}else
					{
						cell->Gp[i]+=tot;
					}

					cell->Gn[i+1]+=tot;
					tot=0.0;
					stored=FALSE;
				}
			}
		}
	}

	for (i=0;i<in->points;i++)
	{
		if (in->enable_gen[in->layer[i]]==0)
		{
			in->Gn[i]=0.0;
			in->Gp[i]=0.0;
		}

	}

	for (i=0;i<in->layers;i++)
	{
		in->G_percent[i]=0.0;
	}

	tot=0.0;
	for (i=0;i<in->points;i++)
	{
		in->G_percent[in->layer[i]]+=in->Gn[i];
		tot+=in->Gn[i];
	}

	for (i=0;i<in->layers;i++)
	{
		in->G_percent[i]*=100.0/tot;
		if (get_dump_status(dump_optics)==TRUE)
		{
			if (tot!=0.0) printf("%d %lf\n",i,in->G_percent[i]);
		}
	}

	light_dump(in);

	if (get_dump_status(dump_optics)==TRUE)
	{
		FILE *out;
		out=fopena(in->output_path,"light_device_G.dat","w");
		for (i=0;i<cell->ymeshpoints;i++)
		{
			fprintf(out,"%le %le %le\n",cell->ymesh[i],cell->Gn[i],cell->Gp[i]);
		}
		fclose(out);
	}

	if (in->inverted==TRUE)
	{

		double *Gn=(double *)malloc(sizeof(double)*cell->ymeshpoints);
		double *Gp=(double *)malloc(sizeof(double)*cell->ymeshpoints);

		for (i=0;i<cell->ymeshpoints;i++)
		{
			Gn[i]=cell->Gn[i];
			Gp[i]=cell->Gp[i];
			//printf("%d %le\n",i,cell->Gn[i]);
		}

		for (i=0;i<cell->ymeshpoints;i++)
		{
			cell->Gn[i]=Gn[cell->ymeshpoints-i-1];
			cell->Gp[i]=Gp[cell->ymeshpoints-i-1];
			//printf("%d %le\n",i,cell->Gn[i]);
		}

		free(Gn);
		free(Gp);
	}


}
//for (i=0;i<cell->ymeshpoints;i++)
//{
//	printf("%le %le %le\n",cell->ymesh[i],in->Gn[i],in->Gp[i]);
//}

//if (cell->Psun!=0) exit(0);

}


int light_solve_lam_slice(struct light *in,int lam)
{
if (get_dump_status(dump_optics)==TRUE)
{
	char one[100];
	sprintf(one,"Solve light optical slice at %lf nm\n",in->l[lam]*1e9);
	//printf("%s\n",one);
	waveprint(one,in->l[lam]*1e9);
}
int i;

double complex Epl=0.0+0.0*I;
double complex Epc=0.0+0.0*I;
double complex Epr=0.0+0.0*I;

double complex Enl=0.0+0.0*I;
double complex Enc=0.0+0.0*I;
double complex Enr=0.0+0.0*I;

double complex r0=0.0+0.0*I;
double complex r1=0.0+0.0*I;
double complex t0=0.0+0.0*I;
double complex t1=0.0+0.0*I;

double yl=0.0;
double yc=0.0;
double yr=0.0;
double dyl=0.0;
double dyr=0.0;
double dyc=0.0;
double xl=0.0;
double xc=0.0;
double xr=0.0;

int ittr=0;
int pos=0;
double error=1000;
double l=in->l[lam];
double nc;
double alpha;
double tot_error=0.0;
int quit=TRUE;

do
{

	pos=0;

	for (i=0;i<in->points;i++)
	{
		if (i==0)
		{
			Epl=in->sun_E[lam]+0.0*I;
			Enl=in->sun_E[lam]+0.0*I;		//don't care
			xl=in->x[i]-(in->x[i+1]-in->x[i]);

		}else
		{
			Epl=in->Ep[lam][i-1]+in->Epz[lam][i-1]*I;
			Enl=in->En[lam][i-1]+in->Enz[lam][i-1]*I;
			xl=in->x[i-1];


		}

		if (i==(in->points-1))
		{
			Epr=1.0+0.0*I;
			Enr=1.0+0.0*I;
			xr=in->x[i]+(in->x[i]-in->x[i-1]);
			r1=in->r[lam][i];
			t1=in->t[lam][i];
		}else
		{
			Epr=in->Ep[lam][i+1]+in->Epz[lam][i+1]*I;
			Enr=in->En[lam][i+1]+in->Enz[lam][i+1]*I;
			xr=in->x[i+1];
			r1=in->r[lam][i+1];
			t1=in->t[lam][i+1];
		}

		xc=in->x[i];
		dyl=xc-xl;
		dyr=xr-xc;
		dyc=(dyl+dyr)/2.0;
		nc=in->n[lam][i];
		alpha=in->alpha[lam][i];
		Epc=in->Ep[lam][i]+in->Epz[lam][i]*I;
		Enc=in->En[lam][i]+in->Enz[lam][i]*I;
		r0=in->r[lam][i];
		t0=in->t[lam][i];
		double complex nbar=((2*pi)/l)*nc+alpha*I;


		//printf("here t0=(%le+i%le)\n",creal(t0),cimag(t0));
	//printf("t=(%le+i%le) n0=(%le+i%le) n1=(%le+i%le)\n",creal(in->r[i][ii]),cimag(in->r[i][ii]),creal(n0),cimag(n0),creal(n1),cimag(n1));
		//printf("r1=(%le+i%le)\n",creal(r1),cimag(r1));
		//printf("t0=(%le+i%le)\n",creal(t0),cimag(t0));
		//printf("t1=(%le+i%le)\n",creal(t1),cimag(t1));

		//getchar();


		double complex fp=0.0;
		double complex fn=0.0;

		double complex pa=t0;//Epl
		double complex pb=-cexp(-nbar*dyl*I);//Epc*
		double complex pc=0.0;
		double complex pbn=-r0*cexp(nbar*dyl*I);

		double complex na=0.0;
		double complex nb=t1;//Enc
		double complex nc=-cexp(nbar*dyl*I);//Enr*
		double complex nbp=-r1*cexp(-nbar*dyl*I);


		if (i==0)
		{
			fp=-Epl;
			pa=0.0;
		}else
		{
			fp=0.0+0.0*I;
		}

		if (i==(in->points-1))
		{
			fn=0.0;//Enr*cexp(nbar*dyl*I);
			nb=1.0;
			nc=1.0;
		}else
		{
			fn=0.0;
		}

		//printf("dy=%le\n",dyl);
		//printf("a(%le+i%le) b(%le+i%le) c(%le+i%le) = %le\n",creal(pa),cimag(pa),creal(pb),cimag(pb),creal(pc),cimag(pc),creal(fp),cimag(fp));
		if (ittr!=0)
		{
			double complex fEp=Epl*t0-Epc*cexp(-nbar*dyl*I)-r0*Enc*cexp(nbar*dyl*I);
			double complex fEn=Enc*t1-Epr*r1*cexp(-nbar*dyl*I)-Enr*cexp(nbar*dyl*I);
			if (i==in->points-1) fEn=Epc+Enc;
			tot_error+=fabs(creal(fEp))+fabs(cimag(fEp))+fabs(creal(fEn))+fabs(cimag(fEn));
			//printf("%d %le errors=fEp(%le+i%le) fEn(%le+i%le) \n",i,tot_error,creal(fEp),cimag(fEp),creal(fEn),cimag(fEn));
			//printf("errors fEp %le \n",);
			//getchar();
		}

		if (i!=0)///////////
		{
			in->Ti[pos]=i;
			in->Tj[pos]=i-1;
			in->Tx[pos]=creal(pa);
			in->Txz[pos]=cimag(pa);
			pos++;
		}

		in->Ti[pos]=i;/////////
		in->Tj[pos]=i;
		in->Tx[pos]=creal(pb);
		in->Txz[pos]=cimag(pb);
		pos++;


		in->Ti[pos]=in->points+i;///////////
		in->Tj[pos]=in->points+i;
		in->Tx[pos]=creal(nb);
		in->Txz[pos]=cimag(nb);
		pos++;

		in->Ti[pos]=i;/////////
		in->Tj[pos]=in->points+i;
		in->Tx[pos]=creal(pbn);
		in->Txz[pos]=cimag(pbn);
		pos++;

		if (i!=in->points-1)//////////
		{
			in->Ti[pos]=in->points+i;
			in->Tj[pos]=i+1;
			in->Tx[pos]=creal(nbp);
			in->Txz[pos]=cimag(nbp);
			pos++;
		}

		if (i!=(in->points-1))///////////////
		{
			in->Ti[pos]=in->points+i;
			in->Tj[pos]=in->points+i+1;
			in->Tx[pos]=creal(nc);
			in->Txz[pos]=cimag(nc);
			pos++;
		}else/////////////////
		{
			in->Ti[pos]=in->points+i;
			in->Tj[pos]=i;
			in->Tx[pos]=creal(nc);
			in->Txz[pos]=cimag(nc);
			pos++;
		}

		in->b[i]=creal(fp);
		in->bz[i]=cimag(fp);

		in->b[in->points+i]=creal(fn);
		in->bz[in->points+i]=cimag(fn);

		//printf("%d %le %le %d %d\n",i,in->b[i],in->b[i+in->points],pos,in->N);

	}

	//printf("%d %d %d\n",in->N,pos,in->M);
	//complex_solver_print(in->M,in->N,in->Ti,in->Tj, in->Tx, in->Txz,in->b,in->bz);
	//getchar();

	if (pos!=in->N)
	{
		printf("not solving because number does not match%d %d\n",pos,in->N);
		exit(0);
	}

	if (lock_detect_gdb()==FALSE) complex_solver(in->M,in->N,in->Ti,in->Tj, in->Tx, in->Txz,in->b,in->bz);

	for (i=0;i<in->points;i++)
	{
		double update;

		update=in->b[i];
		in->Ep[lam][i]=update;

		update=in->bz[i];
		in->Epz[lam][i]=update;

		//printf("b=%le+i%le\n",in->b[i],in->bz[i]);

		update=in->b[in->points+i];
		in->En[lam][i]=update;

		update=in->bz[in->points+i];
		in->Enz[lam][i]=update;

	}

	ittr++;
light_dump_1d(in, lam);
tot_error=0.0;
quit=TRUE;
}while(quit==FALSE);

if (tot_error>1e-10)
{
printf("Error in optical model tot error=%le\n",tot_error);
}
return 0;
}
